import csv
import os


def get_abs_path(rel_path):
    """
    Get absolute path

    Args:
        rel_path: relative path to this file
        
    Returns absolute path
    """
    abs_path = os.path.dirname(os.path.abspath(__file__)) + os.sep + rel_path

    if not os.path.exists(abs_path):
        print(f'{abs_path} does not exist')
    return abs_path


def load_labels(abs_path):
    """
    loads relative path file as dictionary

    Args:
        abs_path: absolute path

    Returns dictionary of mappings
    """
    label_tsv = open(abs_path)
    labels = list(csv.reader(label_tsv, delimiter="\t"))
    return labels



import pynini
from pynini.lib import pynutil

from ukr.graph_utils import GraphFst, NEMO_NOT_SPACE


class WordFst(GraphFst):

    def __init__(self):
        super().__init__(name="word", kind="classify")
        word = pynutil.insert("name: \"") + pynini.closure(NEMO_NOT_SPACE, 1) + pynutil.insert("\"")
        word = self.add_tokens(word)
        self.fst = word.optimize()



from collections import defaultdict

import pynini
from pynini.lib import pynutil

from ukr.graph_utils import GraphFst, delete_space, NEMO_DIGIT
from ukr.taggers.cardinal import CardinalFst
from ukr.utils import get_abs_path, load_labels


def prepare_labels_for_insertion(file_path: str) -> dict:
    labels = load_labels(file_path)
    mapping = defaultdict(list)
    for k, v in labels:
        mapping[k].append(v)

    for k in mapping:
        mapping[k] = pynini.union(*mapping[k]).optimize()
    return mapping


class DecimalFst(GraphFst):

    def __init__(self, cardinal: CardinalFst, ):
        super().__init__(name="decimal", kind="classify")

        delimiter = pynini.string_file(get_abs_path("data/numbers/decimal_delimiter.tsv"))

        delimiter = pynini.cross(delimiter, " ") + pynini.closure(delete_space + pynutil.delete("і"), 0, 1)
        delimiter |= pynini.closure(delete_space + pynini.cross("і", " "), 1, 1)

        decimal_endings_map = prepare_labels_for_insertion(get_abs_path("data/numbers/decimal_endings.tsv"))

        tenth_labels = pynutil.delete(decimal_endings_map["10"])
        tenth = cardinal.graph @ NEMO_DIGIT + delete_space + tenth_labels

        hundreds_labels = pynutil.delete(decimal_endings_map["100"])
        hundreds = cardinal.graph @ (NEMO_DIGIT + NEMO_DIGIT) + delete_space + hundreds_labels
        hundreds |= cardinal.graph @ (pynutil.insert("0") + NEMO_DIGIT) + delete_space + hundreds_labels

        thousands_labels = pynutil.delete(decimal_endings_map["1000"])
        thousands = cardinal.graph @ (NEMO_DIGIT + NEMO_DIGIT + NEMO_DIGIT) + delete_space + thousands_labels
        thousands |= cardinal.graph @ (pynutil.insert("0") + NEMO_DIGIT + NEMO_DIGIT) + delete_space + thousands_labels
        thousands |= cardinal.graph @ (pynutil.insert("0") + pynutil.insert("0") + NEMO_DIGIT) + delete_space + thousands_labels

        graph_fractional_part = pynini.union(
            tenth,
            hundreds,
            thousands
        ).optimize()

        graph_integer = cardinal.graph_integer

        billion = pynini.string_file(get_abs_path("data/numbers/cardinals_billion.tsv"))
        million = pynini.string_file(get_abs_path("data/numbers/cardinals_million.tsv"))

        quantity = pynutil.insert(" quantity: \"") + (billion | million) + pynutil.insert("\"")
        optional_graph_quantity = pynini.closure(" " + quantity, 0, 1)

        graph_fractional = pynutil.insert("fractional_part: \"") + graph_fractional_part + pynutil.insert("\"")

        self.graph = graph_integer + delete_space + delimiter + delete_space + graph_fractional + optional_graph_quantity
        self.graph |= graph_integer + delete_space + quantity
        self.graph |= pynutil.insert("integer_part: \"0\" fractional_part: \"") + (tenth | hundreds) + pynutil.insert("\"")

        optional_minus_graph = pynini.closure(pynutil.insert("negative: \"true\" ") + pynutil.delete("мінус"), 0, 1)

        final_graph = optional_minus_graph + self.graph
        final_graph = self.add_tokens(final_graph)
        self.fst = final_graph.optimize()



import pynini
from pynini.lib import pynutil

from ukr.graph_utils import (
    GraphFst,
    NEMO_CHAR,
    NEMO_DIGIT,
)

from ukr.taggers.cardinal import CardinalFst
from ukr.utils import get_abs_path


class OrdinalFst(GraphFst):

    def __init__(self, cardinal: CardinalFst):
        super().__init__(name="ordinal", kind="classify")

        delete_space = pynutil.delete(" ")
        delete_space_optional = pynini.closure(delete_space, 0, 1)

        graph_cardinal_ties = cardinal.graph_ties
        graph_cardinal_hundred = cardinal.graph_hundred

        graph_ordinal_digit = pynini.invert(pynini.string_file(get_abs_path("data/numbers/ordinal/ordinal_digit.tsv")))
        graph_ordinal_teen = pynini.invert(pynini.string_file(get_abs_path("data/numbers/ordinal/ordinal_teen.tsv")))
        graph_ordinal_ties = pynini.invert(pynini.string_file(get_abs_path("data/numbers/ordinal/ordinal_ties.tsv")))
        graph_ordinal_hundred = pynini.invert(pynini.string_file(get_abs_path("data/numbers/ordinal/ordinal_hundred.tsv")))
        graph_ordinal_thousand = pynini.invert(pynini.string_file(get_abs_path("data/numbers/ordinal/ordinal_thousand.tsv")))

        # From 1 to 99
        graph_up_to_hundred_component = pynini.union(
            graph_ordinal_teen,
            pynutil.insert("0") + graph_ordinal_digit,
            graph_ordinal_ties,
            graph_cardinal_ties + delete_space + graph_ordinal_digit,
        )

        graph_hundred_component = graph_cardinal_hundred + delete_space_optional + graph_up_to_hundred_component
        graph_hundred_component |= graph_ordinal_hundred
        graph_hundred_component |= pynutil.insert("0") + graph_up_to_hundred_component

        thousands = pynini.string_file(get_abs_path("data/numbers/cardinals_thousand.tsv"))

        one_thousand = pynini.cross(thousands, '1') + delete_space + graph_hundred_component
        graph_thousands = (
            pynini.union(
                cardinal.graph_hundred_component + delete_space + pynutil.delete(thousands),
                pynutil.insert("000", weight=0.1),
            )
        )

        millions = pynini.string_file(get_abs_path("data/numbers/cardinals_million.tsv"))
        graph_millions = (
            pynini.union(
                cardinal.graph_hundred_component + delete_space + pynutil.delete(millions),
                pynutil.insert("000", weight=0.1),
            )
        )

        billions = pynini.string_file(get_abs_path("data/numbers/cardinals_billion.tsv"))
        graph_billions = (
            pynini.union(
                cardinal.graph_hundred_component + delete_space + pynutil.delete(billions),
                pynutil.insert("000", weight=0.1),
            )
        )

        # Like: "двох тисячний рік" -> "2000-й рік".
        ordinal_thousand = cardinal.graph_hundred_component + delete_space + graph_ordinal_thousand

        # TODO: add other cardinalities like trillions, etc.
        graph = pynini.union(
            graph_billions
            + delete_space_optional
            + graph_millions
            + delete_space_optional
            + graph_thousands
            + delete_space_optional
            + graph_hundred_component,
            one_thousand,
            ordinal_thousand,
            graph_up_to_hundred_component,
        )

        self.graph_zeroth = graph
        self.graph = graph @ (
                pynini.closure(pynutil.delete(pynini.union("0")))
                + pynini.difference(NEMO_DIGIT, "0")
                + pynini.closure(NEMO_DIGIT)
                + "-" + pynini.closure(NEMO_CHAR)
                + pynini.closure(pynutil.delete(pynini.union("0")))
        )

        # When we have a single digit, like: first, second, ... - leave it as is
        final_graph = self.graph @ (pynini.closure(NEMO_DIGIT, 2) + "-" + pynini.closure(NEMO_CHAR))

        final_graph = pynutil.insert("integer: \"") + final_graph + pynutil.insert("\"")
        final_graph = self.add_tokens(final_graph)

        self.fst = final_graph



import argparse
import sys

from ukr.wfst import graph, json_graph, apply_fst_text

parser = argparse.ArgumentParser()
parser.add_argument('-i', '--inverse', action='store_true', help='inverse inverse text normalization')
parser.add_argument('-j', '--json', action='store_true', help='inverse inverse text normalization')
args = parser.parse_args()

if args.inverse:
    graph = graph.invert()

main_graph = json_graph if args.json else graph

for line in sys.stdin:
    print(apply_fst_text(line.strip(), main_graph))



import pynini
from pynini.lib import pynutil

from ukr.graph_utils import (
    NEMO_DIGIT,
    GraphFst,
)

from ukr.utils import get_abs_path


def get_zeros():
    graph_zeros = pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/nominative/cardinals_zero.tsv")))
    graph_zeros |= pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/genitive/cardinals_zero.tsv")))
    graph_zeros |= pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/dative/cardinals_zero.tsv")))
    graph_zeros |= pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/accusative/cardinals_zero.tsv")))
    graph_zeros |= pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/instrumental/cardinals_zero.tsv")))
    graph_zeros |= pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/prepositional/cardinals_zero.tsv")))

    return graph_zeros.optimize()


def get_digits():
    graph_digit = pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/nominative/cardinals_digit.tsv")))
    graph_digit |= pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/genitive/cardinals_digit.tsv")))
    graph_digit |= pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/dative/cardinals_digit.tsv")))
    graph_digit |= pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/accusative/cardinals_digit.tsv")))
    graph_digit |= pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/instrumental/cardinals_digit.tsv")))
    graph_digit |= pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/prepositional/cardinals_digit.tsv")))

    return graph_digit.optimize()


def get_teen():
    graph_teen = pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/nominative/cardinals_teen.tsv")))
    graph_teen |= pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/genitive/cardinals_teen.tsv")))
    graph_teen |= pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/dative/cardinals_teen.tsv")))
    graph_teen |= pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/accusative/cardinals_teen.tsv")))
    graph_teen |= pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/instrumental/cardinals_teen.tsv")))
    graph_teen |= pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/prepositional/cardinals_teen.tsv")))

    return graph_teen.optimize()


def get_ties():
    graph_ties = pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/nominative/cardinals_ties.tsv")))
    graph_ties |= pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/genitive/cardinals_ties.tsv")))
    graph_ties |= pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/dative/cardinals_ties.tsv")))
    graph_ties |= pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/accusative/cardinals_ties.tsv")))
    graph_ties |= pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/instrumental/cardinals_ties.tsv")))
    graph_ties |= pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/prepositional/cardinals_ties.tsv")))

    return graph_ties.optimize()


def get_hundred():
    graph_ties = pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/nominative/cardinals_hundred.tsv")))
    graph_ties |= pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/genitive/cardinals_hundred.tsv")))
    graph_ties |= pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/dative/cardinals_hundred.tsv")))
    graph_ties |= pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/accusative/cardinals_hundred.tsv")))
    graph_ties |= pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/instrumental/cardinals_hundred.tsv")))
    graph_ties |= pynini.invert(pynini.string_file(get_abs_path("data/numbers/cardinal/prepositional/cardinals_hundred.tsv")))

    return graph_ties.optimize()


class CardinalFst(GraphFst):

    def __init__(self):
        super().__init__(name="cardinal", kind="classify")

        delete_space = pynutil.delete(" ")
        delete_space_optional = pynini.closure(delete_space, 0, 1)

        self.graph_zero = get_zeros()
        self.graph_digit = get_digits()
        self.graph_teen = get_teen()
        self.graph_ties = get_ties()
        self.graph_hundred = get_hundred()

        # From 1 to 99
        graph_up_to_hundred_component = pynini.union(
            self.graph_teen,
            (self.graph_ties | pynutil.insert("0")) + delete_space_optional + (self.graph_digit | pynutil.insert("0"))
        )

        graph_hundred_component = self.graph_hundred + delete_space_optional + graph_up_to_hundred_component
        graph_hundred_component |= pynutil.insert("0") + graph_up_to_hundred_component

        thousands = pynini.string_file(get_abs_path("data/numbers/cardinals_thousand.tsv"))

        one_thousand = pynini.cross(thousands, '1') + delete_space + graph_hundred_component
        graph_thousands = (
            pynini.union(
                graph_hundred_component + delete_space + pynutil.delete(thousands),
                pynutil.insert("000", weight=0.1),
            )
        )

        millions = pynini.string_file(get_abs_path("data/numbers/cardinals_million.tsv"))
        graph_millions = (
            pynini.union(
                graph_hundred_component + delete_space + pynutil.delete(millions),
                pynutil.insert("000", weight=0.1),
            )
        )

        billions = pynini.string_file(get_abs_path("data/numbers/cardinals_billion.tsv"))
        graph_billions = (
            pynini.union(
                graph_hundred_component + delete_space + pynutil.delete(billions),
                pynutil.insert("000", weight=0.1),
            )
        )

        graph = pynini.union(
            graph_billions
            + delete_space_optional
            + graph_millions
            + delete_space_optional
            + graph_thousands
            + delete_space_optional
            + (graph_hundred_component | pynutil.insert("000")),
            one_thousand,
            graph_up_to_hundred_component,
            self.graph_zero,
        )

        self.graph_zeroth = graph
        graph = graph @ pynini.union(
            pynini.closure(pynutil.delete(pynini.union("0", ",")))
            + pynini.difference(NEMO_DIGIT, "0")
            + pynini.closure(pynini.union(NEMO_DIGIT, ",")),
            "0",
        )

        self.graph_hundred_component = graph_hundred_component

        self.graph_up_to_hundred_component = graph_up_to_hundred_component
        self.graph = graph

        self.graph_integer = pynutil.insert("integer_part: \"") + graph + pynutil.insert("\"")

        optional_minus_graph = pynini.closure(pynutil.insert("negative: \"true\" ") + pynutil.delete("мінус"), 0, 1)

        # When we have a single digit, like: one, two, three - leave it as is
        final_graph = graph @ pynini.closure(NEMO_DIGIT, 2)

        final_graph = optional_minus_graph + pynutil.insert("integer: \"") + final_graph + pynutil.insert("\"")
        final_graph = self.add_tokens(final_graph)
        self.fst = final_graph.optimize()



import os
import string
from pathlib import Path

import pynini
from pynini import Far
from pynini.lib import byte, pynutil, utf8

NEMO_CHAR = utf8.VALID_UTF8_CHAR

NEMO_DIGIT = byte.DIGIT
NEMO_LOWER = pynini.union(*string.ascii_lowercase).optimize()
NEMO_UPPER = pynini.union(*string.ascii_uppercase).optimize()
NEMO_ALPHA = pynini.union(NEMO_LOWER, NEMO_UPPER).optimize()
NEMO_ALNUM = pynini.union(NEMO_DIGIT, NEMO_ALPHA).optimize()
NEMO_HEX = pynini.union(*string.hexdigits).optimize()
NEMO_NON_BREAKING_SPACE = u"\u00A0"
NEMO_SPACE = " "
NEMO_WHITE_SPACE = pynini.union(" ", "\t", "\n", "\r", u"\u00A0").optimize()
NEMO_NOT_SPACE = pynini.difference(NEMO_CHAR, NEMO_WHITE_SPACE).optimize()
NEMO_NOT_QUOTE = pynini.difference(NEMO_CHAR, r'"').optimize()

NEMO_PUNCT = pynini.union(*map(pynini.escape, string.punctuation)).optimize()
NEMO_GRAPH = pynini.union(NEMO_ALNUM, NEMO_PUNCT).optimize()

NEMO_SIGMA = pynini.closure(NEMO_CHAR)

delete_space = pynutil.delete(pynini.closure(NEMO_WHITE_SPACE))
delete_zero_or_one_space = pynutil.delete(pynini.closure(NEMO_WHITE_SPACE, 0, 1))
insert_space = pynutil.insert(" ")
delete_extra_space = pynini.cross(pynini.closure(NEMO_WHITE_SPACE, 1), " ")
delete_preserve_order = pynini.closure(
    pynutil.delete(" preserve_order: true")
    | (pynutil.delete(" field_order: \"") + NEMO_NOT_QUOTE + pynutil.delete("\""))
)

TO_LOWER = pynini.union(*[pynini.cross(x, y) for x, y in zip(string.ascii_uppercase, string.ascii_lowercase)])
TO_UPPER = pynini.invert(TO_LOWER)
MIN_NEG_WEIGHT = -0.0001
MIN_POS_WEIGHT = 0.0001


def convert_space(fst) -> 'pynini.FstLike':
    """
    Converts space to nonbreaking space.
    Used only in tagger grammars for transducing token values within quotes, e.g. name: "hello kitty"
    This is making transducer significantly slower, so only use when there could be potential spaces within quotes, otherwise leave it.

    Args:
        fst: input fst

    Returns output fst where breaking spaces are converted to non breaking spaces
    """
    return fst @ pynini.cdrewrite(pynini.cross(NEMO_SPACE, NEMO_NON_BREAKING_SPACE), "", "", NEMO_SIGMA)


class GraphFst:
    """
    Base class for all grammar fsts.

    Args:
        name: name of grammar class
        kind: either 'classify' or 'verbalize'
        deterministic: if True will provide a single transduction option,
            for False multiple transduction are generated (used for audio-based normalization)
    """

    def __init__(self, name: str, kind: str, deterministic: bool = True):
        self.name = name
        self.kind = kind
        self._fst = None
        self.deterministic = deterministic

        self.far_path = Path(os.path.dirname(__file__) + '/grammars/' + kind + '/' + name + '.far')
        if self.far_exist():
            self._fst = Far(self.far_path, mode="r", arc_type="standard", far_type="default").get_fst()

    def far_exist(self) -> bool:
        """
        Returns true if FAR can be loaded
        """
        return self.far_path.exists()

    @property
    def fst(self) -> 'pynini.FstLike':
        return self._fst

    @fst.setter
    def fst(self, fst):
        self._fst = fst

    def add_tokens(self, fst) -> 'pynini.FstLike':
        """
        Wraps class name around to given fst

        Args:
            fst: input fst

        Returns:
            Fst: fst
        """
        return pynutil.insert(f"{self.name} {{ ") + fst + pynutil.insert(" }")

    def delete_tokens(self, fst) -> 'pynini.FstLike':
        """
        Deletes class name wrap around output of given fst

        Args:
            fst: input fst

        Returns:
            Fst: fst
        """
        res = (
                pynutil.delete(f"{self.name}")
                + delete_space
                + pynutil.delete("{")
                + delete_space
                + fst
                + delete_space
                + pynutil.delete("}")
        )
        return res @ pynini.cdrewrite(pynini.cross(u"\u00A0", " "), "", "", NEMO_SIGMA)

    def as_json(self):
        return pynutil.insert('{') \
               + pynutil.insert(f'"{self.name}": ') \
               + pynutil.insert('"') \
               + self.fst \
               + pynutil.insert('"}')



#	Copyright	(c)	2021,	NVIDIA	CORPORATION.		All	rights	reserved.
#
#	Licensed	under	the	Apache	License,	Version	2.0	(the	"License");
#	you	may	not	use	this	file	except	in	compliance	with	the	License.
#	You	may	obtain	a	copy	of	the	License	at
#
#					http://www.apache.org/licenses/LICENSE-2.0
#
#	Unless	required	by	applicable	law	or	agreed	to	in	writing,	software
#	distributed	under	the	License	is	distributed	on	an	"AS	IS"	BASIS,
#	WITHOUT	WARRANTIES	OR	CONDITIONS	OF	ANY	KIND,	either	express	or	implied.
#	See	the	License	for	the	specific	language	governing	permissions	and
#	limitations	under	the	License.



import pynini
from pynini.lib import pynutil

from ukr.graph_utils import GraphFst, delete_space, delete_extra_space
from ukr.verbalizers.verbalize import VerbalizeFst
from ukr.verbalizers.word import WordFst


class VerbalizeFinalFst(GraphFst):
    """
    Finite state transducer that verbalizes an entire sentence, e.g. 
    tokens { name: "its" } tokens { time { hours: "12" minutes: "30" } } tokens { name: "now" } -> its 12:30 now
    """

    def __init__(self):
        super().__init__(name="verbalize_final", kind="verbalize")

        self.verbalize = VerbalizeFst()

        graph = (
                pynutil.delete("tokens")
                + delete_space
                + pynutil.delete("{")
                + delete_space
                + self.verbalize.fst
                + delete_space
                + pynutil.delete("}")
        )
        graph = delete_space + pynini.closure(graph + delete_extra_space) + graph + delete_space

        self.fst = graph

    def as_json(self):
        types = self.verbalize.as_json()
        graph = (
                pynutil.delete("tokens")
                + delete_space
                + pynutil.delete("{")
                + delete_space
                + types
                + delete_space
                + pynutil.delete("}")
        )

        graph = delete_space + pynini.closure(graph + pynutil.insert(",") + delete_extra_space) + graph + delete_space
        graph = pynutil.insert(pynini.escape("[")) + graph + pynutil.insert(pynini.escape("]"))
        return graph



import pynini
from pynini.lib import pynutil

from ukr.graph_utils import delete_space, NEMO_NOT_QUOTE, GraphFst


class DecimalFst(GraphFst):
    """
    Finite state transducer for verbalizing decimal, e.g.
        decimal { negative: "true" integer_part: "12"  fractional_part: "5006" quantity: "billion" } -> -12.5006 billion
    """

    def __init__(self):
        super().__init__(name="decimal", kind="verbalize")
        optionl_sign = pynini.closure(pynini.cross("negative: \"true\"", "-") + delete_space, 0, 1)

        integer = (
                pynutil.delete("integer_part:")
                + delete_space
                + pynutil.delete("\"")
                + pynini.closure(NEMO_NOT_QUOTE, 1)
                + pynutil.delete("\"")
        )
        optional_integer = pynini.closure(integer + delete_space, 0, 1)
        fractional = (
                pynutil.insert(".")
                + pynutil.delete("fractional_part:")
                + delete_space
                + pynutil.delete("\"")
                + pynini.closure(NEMO_NOT_QUOTE, 1)
                + pynutil.delete("\"")
        )
        optional_fractional = pynini.closure(fractional + delete_space, 0, 1)
        quantity = (
                pynutil.delete("quantity:")
                + delete_space
                + pynutil.delete("\"")
                + pynini.closure(NEMO_NOT_QUOTE, 1)
                + pynutil.delete("\"")
        )
        optional_quantity = pynini.closure(pynutil.insert(" ") + quantity + delete_space, 0, 1)
        graph = optional_integer + optional_fractional + optional_quantity
        self.numbers = graph

        graph = optionl_sign + graph
        delete_tokens = self.delete_tokens(graph)
        self.fst = delete_tokens.optimize()



import pynini
from pynini.lib import pynutil

from ukr.graph_utils import GraphFst, delete_space, NEMO_CHAR
from ukr.verbalizers.decimal import DecimalFst


class MoneyFst(GraphFst):
    """
    Finite state transducer for verbalizing money, e.g.
        money { integer_part: "12" fractional_part: "05" currency: "$" } -> $12.05

    Args:
        decimal: DecimalFst
    """

    def __init__(self, decimal: DecimalFst):
        super().__init__(name="money", kind="verbalize")

        units = pynini.union("$", "₴")

        unit = (
                pynutil.delete("currency:")
                + delete_space
                + pynutil.delete("\"")
                + pynini.closure(units, 1)
                + pynutil.delete("\"")
        )
        graph = unit + delete_space + decimal.numbers
        delete_tokens = self.delete_tokens(graph)
        self.fst = delete_tokens.optimize()



import pynini
from pynini.lib import pynutil

from ukr.graph_utils import GraphFst, delete_space, NEMO_CHAR


class MeasureFst(GraphFst):
    """
    Finite state transducer for verbalizing measure, e.g.
        measure { negative: "true" cardinal { integer: "12" } units: "kg" } -> -12 kg

    Args:
        cardinal: CardinalFst
        decimal: DecimalFst
    """

    def __init__(self, cardinal: GraphFst, decimal: GraphFst):
        super().__init__(name="measure", kind="verbalize")
        optional_sign = pynini.closure(pynini.cross("negative: \"true\"", "-"), 0, 1)
        unit = (
                pynutil.delete("units:")
                + delete_space
                + pynutil.delete("\"")
                + pynini.closure(NEMO_CHAR - " ", 1)
                + pynutil.delete("\"")
                + delete_space
        )
        graph_decimal = (
                optional_sign
                + delete_space
                + decimal.numbers
        )
        graph_cardinal = (
                optional_sign
                + delete_space
                + cardinal.numbers

        )
        graph = (graph_cardinal | graph_decimal) + delete_space + pynutil.insert(" ") + unit
        delete_tokens = self.delete_tokens(graph)
        self.fst = delete_tokens.optimize()



import pynini
from pynini.lib import pynutil

from ukr.graph_utils import delete_space, GraphFst, delete_extra_space
from ukr.taggers.cardinal import CardinalFst
from ukr.taggers.date import DateFst
from ukr.taggers.decimal import DecimalFst
from ukr.taggers.measure import MeasureFst
from ukr.taggers.money import MoneyFst
from ukr.taggers.ordinal import OrdinalFst
from ukr.taggers.word import WordFst


class ClassifyFst(GraphFst):

    def __init__(self):
        super().__init__(name="tokenize_and_classify", kind="classify")

        cardinal = CardinalFst()
        cardinal_graph = cardinal.fst

        ordinal = OrdinalFst(cardinal)
        ordinal_graph = ordinal.fst

        decimal = DecimalFst(cardinal)
        decimal_graph = decimal.fst

        measure_graph = MeasureFst(cardinal=cardinal, decimal=decimal).fst
        date_graph = DateFst(cardinal=cardinal, ordinal=ordinal).fst
        word_graph = WordFst().fst
        money_graph = MoneyFst(cardinal=cardinal, decimal=decimal).fst

        classify = (
                pynutil.add_weight(decimal_graph, 1.1)
                | pynutil.add_weight(measure_graph, 1.1)
                | pynutil.add_weight(cardinal_graph, 1.1)
                | pynutil.add_weight(ordinal_graph, 1.1)
                | pynutil.add_weight(money_graph, 1.1)
                | pynutil.add_weight(date_graph, 1.1)
                | pynutil.add_weight(word_graph, 100)
        )

        token = pynutil.insert("tokens { ") + classify + pynutil.insert(" }")

        graph = token + pynini.closure(delete_extra_space + token)
        graph = delete_space + graph + delete_space

        self.fst = graph.optimize()



from ukr.graph_utils import GraphFst
from ukr.verbalizers.cardinal import CardinalFst
from ukr.verbalizers.date import DateFst
from ukr.verbalizers.decimal import DecimalFst
from ukr.verbalizers.measure import MeasureFst
from ukr.verbalizers.money import MoneyFst
from ukr.verbalizers.ordinal import OrdinalFst
from pynini.lib import pynutil

from ukr.verbalizers.word import WordFst


class VerbalizeFst(GraphFst):

    def __init__(self):
        super().__init__(name="verbalize", kind="verbalize")

        self.cardinal = CardinalFst()
        self.decimal = DecimalFst()
        self.ordinal = OrdinalFst()
        self.measure = MeasureFst(decimal=self.decimal, cardinal=self.cardinal)
        self.money = MoneyFst(decimal=self.decimal)
        self.date = DateFst()
        self.word = WordFst()

        graph = (
                self.date.fst
                | self.money.fst
                | self.measure.fst
                | self.ordinal.fst
                | self.decimal.fst
                | self.cardinal.fst
        )
        graph |= pynutil.add_weight(self.word.fst, 100)

        self.fst = graph

    def as_json(self):
        graph = (
                self.money.as_json()
                | self.measure.as_json()
                | self.date.as_json()
                | self.ordinal.as_json()
                | self.decimal.as_json()
                | self.cardinal.as_json()
        )

        graph |= pynutil.add_weight(self.word.as_json(), 100)

        return graph



import pynini

from ukr.taggers.tokenize_and_classify import ClassifyFst
from ukr.verbalizers.verbalize_final import VerbalizeFinalFst


def find_tags(text: str, tagger) -> 'pynini.FstLike':
    """
    Given text use tagger Fst to tag text

    Args:
        text: sentence

    Returns: tagged lattice
    """
    lattice = text @ tagger
    return lattice


def select_tag(lattice: 'pynini.FstLike') -> str:
    """
    Given tagged lattice return shortest path

    Args:
        tagged_text: tagged text

    Returns: shortest path
    """
    tagged_text = pynini.shortestpath(lattice, nshortest=1, unique=True).string()
    return tagged_text


def apply_fst_text(text, fst):
    text = pynini.escape(text)
    tagged_lattice = find_tags(text, fst)
    tagged_text = select_tag(tagged_lattice)

    return tagged_text


classifyFst = ClassifyFst()
verbalizeFinalFst = VerbalizeFinalFst()

graph = pynini.compose(classifyFst.fst, verbalizeFinalFst.fst).optimize()
json_graph = pynini.compose(classifyFst.fst, verbalizeFinalFst.as_json()).optimize()


import pynini
from pynini.lib import pynutil

from ukr.graph_utils import GraphFst, delete_space, NEMO_NOT_QUOTE, NEMO_SIGMA


class OrdinalFst(GraphFst):

    def __init__(self):
        super().__init__(name="ordinal", kind="verbalize")
        graph = (
                pynutil.delete("integer:")
                + delete_space
                + pynutil.delete("\"")
                + pynini.closure(NEMO_NOT_QUOTE, 1)
                + pynutil.delete("\"")
        )

        delete_tokens = self.delete_tokens(graph)
        self.fst = delete_tokens.optimize()



import pynini
from pynini.lib import pynutil

from ukr.graph_utils import GraphFst, delete_space, NEMO_NOT_QUOTE


class CardinalFst(GraphFst):
    """
    Finite state transducer for verbalizing cardinal
        e.g. cardinal { integer: "23" negative: "-" } -> -23
    """

    def __init__(self):
        super().__init__(name="cardinal", kind="verbalize")

        # 'cardinal { negative: "true" integer: "70" }'
        optional_sign = pynini.closure(
            pynutil.delete("negative:")
            + delete_space
            + pynutil.delete("\"")
            + pynini.cross("true", "-")
            + pynutil.delete("\"")
            + delete_space,
            0,
            1,
        )
        graph = (
                pynutil.delete("integer:")
                + delete_space
                + pynutil.delete("\"")
                + pynini.closure(NEMO_NOT_QUOTE, 1)
                + pynutil.delete("\"")
        )
        graph = optional_sign + graph
        self.numbers = graph

        delete_tokens = self.delete_tokens(graph)
        self.fst = delete_tokens.optimize()



import pynini
from pynini.lib import pynutil

from ukr.graph_utils import (
    GraphFst, delete_extra_space,
)

from ukr.taggers.cardinal import CardinalFst
from ukr.taggers.decimal import DecimalFst
from ukr.utils import get_abs_path


class MeasureFst(GraphFst):

    def __init__(self, cardinal: CardinalFst, decimal: DecimalFst):
        super().__init__(name="measure", kind="classify")

        integer_graph = cardinal.graph_integer
        decimal_graph = decimal.graph

        units = pynini.invert(pynini.string_file(get_abs_path("data/measurements.tsv")))
        units = pynutil.insert("units: \"") + units + pynutil.insert("\"")

        optional_minus_graph = pynini.closure(pynutil.insert("negative: \"true\" ") + pynutil.delete("мінус "), 0, 1)

        graph = optional_minus_graph + integer_graph + delete_extra_space + units
        graph |= optional_minus_graph + decimal_graph + delete_extra_space + units

        final_graph = self.add_tokens(graph)
        self.fst = final_graph.optimize()



import pynini
from pynini.lib import pynutil

from ukr.graph_utils import GraphFst, NEMO_CHAR, delete_space, NEMO_SIGMA


class WordFst(GraphFst):
    """
    Finite state transducer for verbalizing plain tokens
        e.g. tokens { name: "sleep" } -> sleep
    """

    def __init__(self):
        super().__init__(name="word", kind="verbalize")
        chars = pynini.closure(NEMO_CHAR - " ", 1)
        char = pynutil.delete("name:") + delete_space + pynutil.delete("\"") + chars + pynutil.delete("\"")
        graph = char @ pynini.cdrewrite(pynini.cross(u"\u00A0", " "), "", "", NEMO_SIGMA)

        graph = self.delete_tokens(graph)
        self.fst = graph.optimize()



import pynini
from pynini.lib import pynutil

from ukr.graph_utils import GraphFst, delete_space, delete_extra_space, NEMO_CHAR, NEMO_DIGIT
from ukr.taggers.cardinal import CardinalFst
from ukr.taggers.ordinal import OrdinalFst
from ukr.utils import get_abs_path


class DateFst(GraphFst):
    """
    Finite state transducer for classifying date,
        e.g. january fifth twenty twelve -> date { month: "january" day: "5" year: "2012" }

    Args:
        cardinal: CardinalFst
        ordinal: OrdinalFst
    """

    def __init__(self, cardinal: CardinalFst, ordinal: OrdinalFst):
        super().__init__(name="date", kind="classify")

        graph_cardinal_ties = cardinal.graph_ties

        graph_ordinal_digit = pynini.invert(pynini.string_file(get_abs_path("data/numbers/ordinal/ordinal_digit.tsv")))
        graph_ordinal_teen = pynini.invert(pynini.string_file(get_abs_path("data/numbers/ordinal/ordinal_teen.tsv")))
        graph_ordinal_ties = pynini.invert(pynini.string_file(get_abs_path("data/numbers/ordinal/ordinal_ties.tsv")))

        day = pynini.union(
            graph_ordinal_digit,
            graph_ordinal_teen,
            pynutil.insert("0") + graph_ordinal_digit,
            graph_ordinal_ties,
            graph_cardinal_ties + delete_space + graph_ordinal_digit,
        )
        day = day @ (pynini.closure(NEMO_DIGIT) + pynutil.delete(pynini.union("-") + pynini.closure(NEMO_CHAR)))
        graph_day = pynutil.insert("day: \"") + day + pynutil.insert("\"")

        graph_month = pynutil.insert("month: \"") + pynini.string_file(get_abs_path("data/month.tsv")) + pynutil.insert("\"")

        year_name = delete_extra_space + pynini.union("року", "рік")
        optional_year_name = pynini.closure(year_name, 0, 1)

        optional_epoch = delete_extra_space + pynini.closure("до ", 0, 1) + pynini.cross("нашої ери", "н. е.")
        optional_epoch = pynini.closure(optional_epoch, 0, 1)

        year = ordinal.graph @ (pynini.closure(NEMO_DIGIT) + pynutil.delete(pynini.union("-") + pynini.closure(NEMO_CHAR)))

        graph_year = pynutil.insert("year: \"") + year + optional_year_name + optional_epoch + pynutil.insert("\"")
        graph_only_year = pynutil.insert("year: \"") + year + year_name + optional_epoch + pynutil.insert("\"")

        graph_dmy = graph_day + delete_extra_space + graph_month + delete_extra_space + graph_year
        graph_my = graph_month + delete_extra_space + graph_year
        graph_dm = graph_day + delete_extra_space + graph_month

        final_graph = graph_dmy | graph_my | graph_dm | graph_only_year

        final_graph = self.add_tokens(final_graph)
        self.fst = final_graph.optimize()



from collections import defaultdict

import pynini
from pynini.lib import pynutil

from ukr.graph_utils import (
    GraphFst,
    NEMO_DIGIT,
    delete_space,
)

from ukr.taggers.cardinal import CardinalFst
from ukr.taggers.decimal import DecimalFst
from ukr.utils import get_abs_path, load_labels


def load_currency(file_path):
    labels = load_labels(file_path)

    mapping = defaultdict(list)
    for k, v in labels:
        mapping[k].append(v)

    for k in mapping:
        mapping[k] = pynini.union(*mapping[k]).optimize()

    return mapping


class MoneyFst(GraphFst):
    """
    Finite state transducer for classifying money
        e.g. twelve dollars and five cents -> money { integer_part: "12" fractional_part: 05 currency: "$" }

    Args:
        cardinal: CardinalFst
        decimal: DecimalFst
    """

    def __init__(self, cardinal: CardinalFst, decimal: DecimalFst):
        super().__init__(name="money", kind="classify")

        integer_graph = pynutil.insert("integer_part: \"") + cardinal.graph + pynutil.insert("\" ")
        decimal_graph = decimal.graph

        # сім копійок -> 0.07
        # двадцять сім копійок -> 0.27
        add_leading_zero_to_double_digit = (NEMO_DIGIT + NEMO_DIGIT) | (pynutil.insert("0") + NEMO_DIGIT)
        fractional_graph = pynutil.insert("fractional_part: \"") + cardinal.graph @ add_leading_zero_to_double_digit + pynutil.insert("\"")

        unit_majors = load_currency(get_abs_path("data/currency/currency_major.tsv"))
        unit_minors = load_currency(get_abs_path("data/currency/currency_minor.tsv"))

        units = []
        optional_and = pynini.closure(delete_space + pynutil.delete("і"), 0, 1)
        for signature, major in unit_majors.items():
            minor = unit_minors[signature]

            # just integer: one dollar, two dollars
            integer_part = integer_graph + delete_space + pynutil.delete(major)
            units.append(pynutil.insert(f"currency: \"{signature}\" ") + integer_part)

            # just integer and fractional: one dollar and two cents, two dollars two cents, etc.
            fractional_part = fractional_graph + delete_space + pynutil.delete(minor)
            unit = integer_part + optional_and + delete_space + fractional_part
            units.append(pynutil.insert(f"currency: \"{signature}\" ") + unit)

            # just fractional: two cents, one cents, etc.
            integer_part = pynutil.insert("integer_part: \"0\" ")
            fractional_part = fractional_graph + delete_space + pynutil.delete(minor)
            unit = integer_part + fractional_part
            units.append(pynutil.insert(f"currency: \"{signature}\" ") + unit)

            # decimal: two point one million dollars, etc.
            unit = decimal_graph + delete_space + pynutil.delete(major)
            units.append(pynutil.insert(f"currency: \"{signature}\" ") + unit)

        final_graph = pynini.union(*units)
        final_graph = self.add_tokens(final_graph)
        self.fst = final_graph.optimize()



import pynini
from pynini.lib import pynutil

from ukr.graph_utils import GraphFst, delete_space, NEMO_NOT_QUOTE, delete_extra_space


class DateFst(GraphFst):

    def __init__(self):
        super().__init__(name="date", kind="verbalize")
        month = (
                pynutil.delete("month:")
                + delete_space
                + pynutil.delete("\"")
                + pynini.closure(NEMO_NOT_QUOTE, 1)
                + pynutil.delete("\"")
        )
        day = (
                pynutil.delete("day:")
                + delete_space
                + pynutil.delete("\"")
                + pynini.closure(NEMO_NOT_QUOTE, 1)
                + pynutil.delete("\"")
        )
        year = (
                pynutil.delete("year:")
                + delete_space
                + pynutil.delete("\"")
                + pynini.closure(NEMO_NOT_QUOTE, 1)
                + delete_space
                + pynutil.delete("\"")
        )

        graph_dmy = day + delete_extra_space + month + delete_extra_space + year
        graph_my = month + delete_extra_space + year
        graph_dm = day + delete_extra_space + month

        final_graph = graph_dmy | graph_my | graph_dm | year

        delete_tokens = self.delete_tokens(final_graph)
        self.fst = delete_tokens.optimize()
